function [csout] = refine(beh,prob,qmin,fid)
%
% Implementa algorisme de REFINEMENT
%
% Parametres d'entrada:
%
%       beh - Matriu behavior del sistema
%       prob - Vector de probabilitat del sistema
%       qmin - Qualitat minima requerida en la reconstruccio
%       fid - Indica si volem fitxer amb resultats intermitjos
%
% Inicialitza variables
%
[mb,nb] = size(beh);
%
% Troba les relacions binaries del sistema
% i escriu capalera al fitxer
%
cst=1:nb;
bin = binary(cst);
if fid ~= 0
  wriline(fid,1,cst,nb);
  wriheader(fid,1,cst,nb);
end;
%
% Troba l'error de la "totally unconnected structure"
%
cst = zeros(1,(nb*2)-1);
index = 1:2:(nb*2)-1;
for i=1:nb,
 cst(index(i)) = i;
end;
errunc = structure(beh,prob,cst(1:(nb*2)-1),1);
[nbin,b] = size(bin);
qual = zeros(nbin,1);
%
% Bucle que elimina una relacio binaria i comproba quin es l'error
% de reconstruccio sense aquesta relacio binaria
%
relbin = nbin;
while ( relbin > 1 )
  for i=1:relbin,
    binaux = [bin(1:i-1,:);bin(i+1:relbin,:)];
    cst = compose(binaux);
    aux = max(cst);		% Proteccio necessaria per a que no
    if aux ~= nb		% falli la funcio structure en cas
      for j=1:nb-aux,		% de que en les relacions binaries
        cst = [cst,0,aux+j];	% de binaux no estiguin els nombres
      end;			% representant a totes les variables
    end;
    errst = structure(beh,prob,cst,1);
    [a,b] = size(cst);
    qual(i) = 1-(errst/errunc);
    if fid ~= 0 wriline(fid,qual(i),cst,b); end;
  end;
%
% Escull quina es la relacio binaria que menys perdua de qualitat
% dona i eliminar-la. Si aquesta qualitat es > qmin seguim si no sortim
%
  [qmax,k] = max(qual);
  if qmax <= qmin break; end;
  bin = [bin(1:k-1,:);bin(k+1:relbin,:)];  
  relbin = relbin-1;
  if fid ~= 0
    clear cst;
    cst = compose(bin);
    aux = max(cst);		% Proteccio necessaria per escriure
    if aux ~= nb		% correctament al fitxer en cas
      for j=1:nb-aux,		% de que en les relacions binaries
        cst = [cst,0,aux+j];	% de bin no estiguin els nombres
      end;			% representant a totes les variables
    end;
    [a,b] = size(cst);
    wriheader(fid,qual(k),cst,b);
  end;
  for i=1:nbin,
    qual(i) = 0;
  end;
end;
%
% Escriu les ultimes linies al fitxer segons condicio sortida
%
if relbin == 1
  csout = cst;
  if fid ~= 0
    for i=1:nb*2,
      cst(i) = 0;
    end;
    j = 0;
    for i=1:nb,
      cst(i+j) = i;
      j = i;
    end;
    wriline(fid,0,cst,(2*nb)-1);
    fprintf(fid,'\nMILLOR ESTRUCTURA EN AQUEST NIVELL TE QUALITAT 0.0000\n');
    fprintf(fid,'CONSIDERADA MASSA BAIXA. ACABA LA CERCA\n\n');
    fprintf(fid,'ESTRUCTURA RESULTANT: (');
    [a,b]=size(csout);
    cs2 = [csout,0];
    for i=1:b,
      if cs2(i) ~= 0 fprintf(fid,'%d',cs2(i)); end;
      if ( cs2(i+1) ~= 0 & cs2(i) ~= 0 ) fprintf(fid,','); end;
      if cs2(i) == 0 fprintf(fid,') ('); end;
    end;
    fprintf(fid,')\n');
  end;
end;
%
if relbin > 1
  csout = compose(bin);
  aux = max(csout);		% Proteccio necessaria per escriure
  if aux ~= nb			% correctament al fitxer en cas
    for j=1:nb-aux,		% de que en les relacions binaries
      csout = [csout,0,aux+j];	% de bin no estiguin els nombres
    end;			% representant a totes les variables
  end;
  if fid ~= 0
    fprintf(fid,'\nMILLOR ESTRUCTURA EN AQUEST NIVELL TE QUALITAT: %6.4f\n',qual(k));
    fprintf(fid,'CONSIDERADA MASSA BAIXA. ACABA LA CERCA\n\n');
    fprintf(fid,'ESTRUCTURA RESULTANT: ('); 
    [a,b]=size(csout);
    cs2 = [csout,0];
    for i=1:b,
      if cs2(i) ~= 0 fprintf(fid,'%d',cs2(i)); end;
      if ( cs2(i+1) ~= 0 & cs2(i) ~= 0 ) fprintf(fid,','); end;
      if cs2(i) == 0 fprintf(fid,') ('); end;
    end;
    fprintf(fid,')\n');
  end;
end;
return
